//---------------------------------------------------------------- -o--
// RomanNumeral.java
// 
// Convert input Roman Numeral input String to an Arabic Numeral.
// Errors expressed via GTException.
//
// THROWS GTException on error.
//
//
// INSTANCE METHODS--
//   RomanNumeral
//   resolveRomanNumeralValue
//   asArabicNumeral
//
// HELPER METHODS--
//   setRNIndices
//   incrementRNIndicesBy
//
//
// DEPENDS UPON--
//   RomanNumber
//
//
// @version 0.1 
//

import java.util.*;



//-------------------------------------------------------------- -o--
public class RomanNumeral 
{                                                       
  // Properties of RumanNumeral.
  //
  private String  romanNumeral   = null;
  private Float   arabicNumeral  = new Float(0);


  // Indices and state for parse of Roman Numeral Expression into Arabic
  //   Numeral Expression.
  //
  private Integer               rnStrIndex = 0;
  private static final Integer  INDEX_OUTOFRANGE = -1;

  private RomanNumber  rn           = null,  // current RomanNumber in sequence
                       rnNext       = null,  // next RomanNumber in sequence
                       currentRank  = null;


  private final String  ROMANNUMERAL_ERROR = 
				    getClass() + ": Invalid expression -- ";



  //-------------------------------------------------------------- -o--
  // Constructor.
  //
  // THROWS GTException if input is null or contains tokens that are not
  //   known RomanNumbers.
  //

  //---------------- -o-
  RomanNumeral(String romanNumeralExpression)  throws GTException
  { 
    romanNumeral = romanNumeralExpression;
    resolveRomanNumeralValue();
  } 


  //---------------- -o-
  private Float  resolveRomanNumeralValue()  throws GTException
  {
    // Sanity check input.
    // Reject inputs containing non-RomanNumber characters.
    //
    if ((null == romanNumeral) || (0 == romanNumeral.length()))
    {
      throw new GTException(
          ROMANNUMERAL_ERROR + "Input must include one or more RomanNumber's.");
    }

    //
    String  allRomanNumbersRegex = "";

    for (RomanNumber rn : RomanNumber.values()) {
      allRomanNumbersRegex += rn.toString();
    }

    allRomanNumbersRegex = "[" + allRomanNumbersRegex + "]";

    if (0 != romanNumeral.replaceAll(allRomanNumbersRegex, "").length()) 
    {
      throw new GTException(
          ROMANNUMERAL_ERROR + "Contains non-Roman Number characters.");
    }


    // Initialize parse with untallied set of RomanNumbers.
    //
    for (RomanNumber i : RomanNumber.values()) {
      i.resetToDefaults();
    }


    // Parse input.
    //
    setRNIndices(0);
    currentRank = rn;

    while (null != rn) 
    {
      if (! rn.getIsAvailable()) {
        throw new GTException(ROMANNUMERAL_ERROR 
          +"RomanNumeral "+ rn.toString() +" has already been used.");
      }

      if (rn.isGreaterThan(currentRank)) {
        throw new GTException(ROMANNUMERAL_ERROR 
          +"RomanNumeral "+ rn.toString() 
            +" cannot be of higher rank than previous character.");
      }


      //
      if (null == rnNext) 
      {
        arabicNumeral += rn.asArabicNumeral();
        incrementRNIndicesBy(1);


      } else if (rn.equals(rnNext)) 
      {
        rn.addOccurence();
        arabicNumeral += rn.asArabicNumeral();
        incrementRNIndicesBy(1);


      } else if (rn.isLessThan(rnNext)) 
      {
        if (! rn.maySubtractFrom(rnNext)) {
          throw new GTException(ROMANNUMERAL_ERROR 
            +"RomanNumeral "+ rn.toString() 
              +" cannot subtract from "+ rnNext.toString() +".");
        }

        arabicNumeral += (rnNext.asArabicNumeral() - rn.asArabicNumeral());
        rn.setSubtracted();
        if (rnNext.getIsAvailable())  { currentRank = rnNext; }

        // Ordering Roman Numbers: Rule #4
        //
        Iterator<RomanNumber>  rnIter = rn.subtracteeSet.iterator();
        while (rnIter.hasNext()) {
          rnIter.next().setIsAvailable(false);
        }

        incrementRNIndicesBy(2);


      } else  // if (rn.isGreaterThan(rnNext))
      { 
        arabicNumeral += rn.asArabicNumeral();
        rn.addOccurence();
        currentRank = rnNext;
        incrementRNIndicesBy(1);

      } // endif
    } // endwhile


    return arabicNumeral;

  } // resolveRomanNumeralValue()



  //-------------------------------------------------------------- -o--
  // Helper methods.
  //

  //---------------- -o-
  private void  setRNIndices(Integer index)
  {
    if (rnStrIndex <= INDEX_OUTOFRANGE)  { return; }


    // RomanNumbers higher in rank than currentRank are not available.
    //
    if (null != currentRank) {
      for (RomanNumber i: RomanNumber.values()) {
        if (i.isGreaterThan(currentRank)) {
          i.setIsAvailable(false);
        }
      }
    }


    //
    rn = rnNext = null;

    rnStrIndex  = (index >= romanNumeral.length()) ? INDEX_OUTOFRANGE : index;

    index += 1;
    Integer rnStrIndexNext 
                = (index >= romanNumeral.length()) ? INDEX_OUTOFRANGE : index;


    if (rnStrIndex > INDEX_OUTOFRANGE) 
    {
      rn = RomanNumber.valueOf(romanNumeral.charAt(rnStrIndex) + "");

      if (rnStrIndexNext > INDEX_OUTOFRANGE) {
        rnNext = RomanNumber.valueOf(romanNumeral.charAt(rnStrIndexNext) + "");
      }
    }

  } // setRNIndices


  //---------------- -o-
  private void  incrementRNIndicesBy(Integer increment)
  {
    rnStrIndex += increment;
    setRNIndices(rnStrIndex);
  }



  //-------------------------------------------------------------- -o--
  // Instance methods.
  //

  //---------------- -o-
  Float asArabicNumeral()  { return arabicNumeral; }


}  // public class RomanNumeral 

